/*
+----------------+
| GaugeCluster.h |
+----------------+

By Andrew Wyatt
retrojdm.com

Started:		2014-02-01
Updated:		2014-06-12


Last Change:	Renamed Compass to Heading


To Do:

[ ]	Analog Clock: 720 frame (12hr x 60min) GIF instead?

--------------------------------------------------------------------------------
*/


#ifndef GaugeCluster_h
#define GaugeCluster_h

#include <Arduino.h>

#include <avr/eeprom.h>
#include <MemoryFree.h>
#include <SoftwareSerial.h>
#include <Goldelox_Serial_4DLib.h>
#include <Goldelox_const4D.h>

// +----------+
// | #Defines |
// +----------+
#define DWORD unsigned long



// +-----------+
// | Constants |
// +-----------+
static const char *	GAUGE_CLUSTER_VERSION					= "2014-06-12";

// EEPROM Addresses
static const int	SETTINGS_ADDRESS						= 0;
static const char *	SETTINGS_ID_STRING						= "GCL";

static const float	MPH_PER_KNOT							=   1.15077945;
static const float	KMPH_PER_KNOT							=   1.852;


static const int	GAUGE_BUTTON_HOLD						= 500;	// milliseconds

static const byte	MODE_SCREENS							=   0;
static const byte	MODE_MENU								=   1;
static const byte	MODE_DIALOG								=   2;

static const byte	SCREENS									=   8;
static const byte	SCREEN_SPEEDO_MPH						=   0;
static const byte	SCREEN_SPEEDO_KMPH						=   1;
static const byte	SCREEN_SPEEDO_KNOTS						=   2;
static const byte	SCREEN_TACHO							=   3;
static const byte	SCREEN_CLOCK							=   4;
static const byte	SCREEN_HEADING							=   5;
static const byte	SCREEN_GPS_HOME							=   6;
static const byte	SCREEN_GPS_LOCATION						=   7;

static const byte	STYLE_ANALOG							=   0;
static const byte	STYLE_DIGITAL							=   1;

static const byte	DISPLAY_WIDTH							= 160;	// px
static const byte	DISPLAY_HEIGHT							= 128;	// px

static const byte	MENU_TITLE_TOP							=   6;	// px

static const byte	MENU_ARROW_LEFT							=  74;	// px
static const byte	MENU_ARROW_TOP							=  16;	// px
static const byte	MENU_ARROW_BOTTOM						= 122;	// px

static const byte	MENU_ITEMS_TOP							=  24;	// px
static const byte	MENU_ITEM_HEIGHT						=  24;	// px
static const byte	MENU_ITEM_PADDING_LEFT					=  24;	// px
static const byte	MENU_ITEM_BASELINE						=   9;	// px

static const byte	MENU_ICON_PADDING_LEFT					=   2;	// px
static const byte	MENU_ICON_PADDING_TOP					=   2;	// px

static const byte	MENU_CHECKBOX_PADDING_LEFT				=   6;	// px
static const byte	MENU_CHECKBOX_PADDING_TOP				=   6;	// px

static const byte	MENU_RADIO_PADDING_LEFT					=   6;	// px
static const byte	MENU_RADIO_PADDING_TOP					=   6;	// px

// Item types
static const byte	TYPE_ITEM								=   1;
static const byte	TYPE_CHECKBOX							=   2;
static const byte	TYPE_RADIOBUTTON						=   3;

// The gauge cluster has an array of settings.
// Instead of passing a pointer of to the variable we want the checkbox or radio button to alter,
// we send it the index of the setting.
// This is because we don't know ahead of time what the address of the variable will be.

static const byte	NONE									=   0;

static const byte	CLUSTER_SETTINGS						=   3;
static const byte	CLUSTER_STYLE							=   0;
static const byte	CLUSTER_TIMEZONE						=   1;
static const byte	CLUSTER_COLOUR							=   2;

// It's important that these match the screens order
static const byte	GAUGE_SPEEDO_MPH						=   3;
static const byte	GAUGE_SPEEDO_KMPH						=   4;
static const byte	GAUGE_SPEEDO_KNOTS						=   5;
static const byte	GAUGE_TACHO								=   6;
static const byte	GAUGE_CLOCK								=   7;
static const byte	GAUGE_HEADING							=   8;
static const byte	GAUGE_GPS_HOME							=   9;
static const byte	GAUGE_GPS_LOCATION						=  10;
static const byte	GAUGE_CURRENT_SCREEN					=  11;	//	Which screen are we currently on?

// Commands
static const byte	CMD_EXIT								=   1;
static const byte	CMD_SET_GPS								=   2;
static const byte	CMD_ABOUT								=   3;
static const byte	CMD_SET_MENU							=   4;

// +----------------------+
// | uSD Images Addresses |
// +----------------------+

// RetroJDM Logo
static const DWORD	IMAGE_RETROJDM							= 0x00000000;

// Controls
static const DWORD	IMAGE_CONTROL_ARROWDOWN					= 0x00001400;
static const DWORD	IMAGE_CONTROL_ARROWUP					= 0x00001600;
static const DWORD	IMAGE_CONTROL_CHECKBOX					= 0x00001800;
static const DWORD	IMAGE_CONTROL_CHECKBOX_CHECKED			= 0x00001A00;
static const DWORD	IMAGE_CONTROL_CHECKBOX_HIGHLIGHT		= 0x00001C00;
static const DWORD	IMAGE_CONTROL_CHECKBOX_HIGHLIGHT_CHECKED= 0x00001E00;
static const DWORD	IMAGE_CONTROL_RADIO						= 0x00002000;
static const DWORD	IMAGE_CONTROL_RADIO_CHECKED				= 0x00002200;
static const DWORD	IMAGE_CONTROL_RADIO_HIGHLIGHT			= 0x00002400;
static const DWORD	IMAGE_CONTROL_RADIO_HIGHLIGHT_CHECKED	= 0x00002600;

// Icons
static const word	ICON_HIGHLIGHT_OFFSET					= 0x0400;
static const DWORD	IMAGE_ICON_COLOURS						= 0x00002800;
static const DWORD	IMAGE_ICON_EXIT							= 0x00003000;
static const DWORD	IMAGE_ICON_GAUGEANALOG					= 0x00003800;
static const DWORD	IMAGE_ICON_GAUGEDIGITAL					= 0x00004000;
static const DWORD	IMAGE_ICON_GAUGEREDLINE					= 0x00004800;
static const DWORD	IMAGE_ICON_GAUGESTYLE					= 0x00005000;
static const DWORD	IMAGE_ICON_GLOBE						= 0x00005800;
static const DWORD	IMAGE_ICON_HOMEPIN						= 0x00006000;
static const DWORD	IMAGE_ICON_INFO							= 0x00006800;
static const DWORD	IMAGE_ICON_OPTIONS						= 0x00007000;
static const DWORD	IMAGE_ICON_SHIFTER						= 0x00007800;

// Videos (GIFs)
static const DWORD	VIDEO_SPEEDO_ANALOG_MPH					= 0x000B7400;
static const byte	VIDEO_SPEEDO_ANALOG_MPH_PER_FRAME		= 1;
static const byte	VIDEO_SPEEDO_ANALOG_MPH_LAST_FRAME		= 80;

static const DWORD	VIDEO_SPEEDO_ANALOG_KMPH				= 0x003E1600;
static const byte	VIDEO_SPEEDO_ANALOG_KMPH_PER_FRAME		= 1;
static const byte	VIDEO_SPEEDO_ANALOG_KMPH_LAST_FRAME		= 160;

static const DWORD	VIDEO_SPEEDO_ANALOG_KNOTS				= 0x00A2B800;
static const byte	VIDEO_SPEEDO_ANALOG_KNOTS_PER_FRAME		= 1;
static const byte	VIDEO_SPEEDO_ANALOG_KNOTS_LAST_FRAME	= 80;

static const DWORD	VIDEO_TACHO_ANALOG						= 0x00D55A00;
static const byte	VIDEO_TACHO_ANALOG_RPM_PER_FRAME		= 100;
static const byte	VIDEO_TACHO_ANALOG_LAST_FRAME			= 80;

static const DWORD	IMAGE_CLOCK								= 0x0107FC00;

static const DWORD	VIDEO_HEADING_DIGITAL					= 0x01089E00;
static const DWORD	VIDEO_HEADING_ANALOG					= 0x010BE000;

static const DWORD	VIDEO_GPS_HOME_DIGITAL					= 0x01ECE200;
static const DWORD	VIDEO_GPS_HOME_ANALOG					= 0x0276B400;

static const DWORD	VIDEO_SATELLITE_FIX						= 0x03008600;
static const DWORD	VIDEO_SATELLITE_FIX_FRAMES				= 6;
static const DWORD	VIDEO_SATELLITE_FIX_FRAME_DELAY			= 500; // milliseconds


// Analog Clock
static const byte	HOUR_HAND_LENGTH						=  40;	// px (clock face radius = 64px)
static const byte	MINUTE_HAND_LENGTH						=  64;	// px



// +------------+
// | Menu Items |
// +------------+
static const byte	MAX_ITEMS_PER_MENU						=   31;
static const byte	ITEMS_PER_SCREEN						=    4;
static const byte	ITEMCOUNT								=	59;

typedef struct MenuItemType {
	byte	parent;
	byte	type;
	char *	name;
	DWORD	icon;		// Address of raw data on uSD.
	byte	command;
	word	value;
	byte	setting;
};

static const MenuItemType menuItems[ITEMCOUNT] = {
	{	0,	TYPE_ITEM,			"Exit",				IMAGE_ICON_EXIT,		CMD_EXIT,		NONE,			NONE				},	//  0
	{	0,	TYPE_ITEM,			"Set GPS Home",		IMAGE_ICON_HOMEPIN,		CMD_SET_GPS,	NONE,			NONE				},	//  1
	{	0,	TYPE_ITEM,			"About",			IMAGE_ICON_INFO,		CMD_ABOUT,		NONE,			NONE				},	//  2
	{	0,	TYPE_ITEM,			"Settings",			IMAGE_ICON_SHIFTER,		CMD_SET_MENU,	   3,			NONE				},	//  3
	{	3,	TYPE_ITEM,			"Exit",				IMAGE_ICON_EXIT,		CMD_EXIT,		NONE,			NONE				},	//  4
	{	3,	TYPE_ITEM,			"Set Timezone",		IMAGE_ICON_GLOBE,		CMD_SET_MENU,	   5,			NONE				},	//  5
	{	3,	TYPE_ITEM,			"Screens",			IMAGE_ICON_OPTIONS,		CMD_SET_MENU,	   6,			NONE				},	//  6
	{	3,	TYPE_ITEM,			"Gauge Style",		IMAGE_ICON_GAUGESTYLE,	CMD_SET_MENU,	   7,			NONE				},	//  7
	{	3,	TYPE_ITEM,			"Heading Colour",	IMAGE_ICON_COLOURS,		CMD_SET_MENU,	   8,			NONE				},	//  8
	{	5,	TYPE_ITEM,			"Exit",				IMAGE_ICON_EXIT,		CMD_EXIT,		NONE,			NONE				},	//  9
	{	5,	TYPE_RADIOBUTTON,	"-12",				NONE,					NONE,			-720,			CLUSTER_TIMEZONE	},	// 10
	{	5,	TYPE_RADIOBUTTON,	"-11",				NONE,					NONE,			-660,			CLUSTER_TIMEZONE	},	// 11
	{	5,	TYPE_RADIOBUTTON,	"-10",				NONE,					NONE,			-600,			CLUSTER_TIMEZONE	},	// 12
	{	5,	TYPE_RADIOBUTTON,	" -9",				NONE,					NONE,			-540,			CLUSTER_TIMEZONE	},	// 13
	{	5,	TYPE_RADIOBUTTON,	" -8",				NONE,					NONE,			-480,			CLUSTER_TIMEZONE	},	// 14
	{	5,	TYPE_RADIOBUTTON,	" -7",				NONE,					NONE,			-420,			CLUSTER_TIMEZONE	},	// 15
	{	5,	TYPE_RADIOBUTTON,	" -6",				NONE,					NONE,			-360,			CLUSTER_TIMEZONE	},	// 16
	{	5,	TYPE_RADIOBUTTON,	" -5",				NONE,					NONE,			-300,			CLUSTER_TIMEZONE	},	// 17
	{	5,	TYPE_RADIOBUTTON,	" -4",				NONE,					NONE,			-240,			CLUSTER_TIMEZONE	},	// 18
	{	5,	TYPE_RADIOBUTTON,	" -3.5",			NONE,					NONE,			-210,			CLUSTER_TIMEZONE	},	// 19
	{	5,	TYPE_RADIOBUTTON,	" -3",				NONE,					NONE,			-180,			CLUSTER_TIMEZONE	},	// 20
	{	5,	TYPE_RADIOBUTTON,	" -2",				NONE,					NONE,			-120,			CLUSTER_TIMEZONE	},	// 21
	{	5,	TYPE_RADIOBUTTON,	" -1",				NONE,					NONE,			 -60,			CLUSTER_TIMEZONE	},	// 22
	{	5,	TYPE_RADIOBUTTON,	"  0 (GMT)",		NONE,					NONE,			   0,			CLUSTER_TIMEZONE	},	// 23
	{	5,	TYPE_RADIOBUTTON,	" +1",				NONE,					NONE,			  60,			CLUSTER_TIMEZONE	},	// 24
	{	5,	TYPE_RADIOBUTTON,	" +2",				NONE,					NONE,			 120,			CLUSTER_TIMEZONE	},	// 25
	{	5,	TYPE_RADIOBUTTON,	" +3",				NONE,					NONE,			 180,			CLUSTER_TIMEZONE	},	// 26
	{	5,	TYPE_RADIOBUTTON,	" +3.5",			NONE,					NONE,			 210,			CLUSTER_TIMEZONE	},	// 27
	{	5,	TYPE_RADIOBUTTON,	" +4",				NONE,					NONE,			 240,			CLUSTER_TIMEZONE	},	// 28
	{	5,	TYPE_RADIOBUTTON,	" +4.5",			NONE,					NONE,			 270,			CLUSTER_TIMEZONE	},	// 29
	{	5,	TYPE_RADIOBUTTON,	" +5",				NONE,					NONE,			 300,			CLUSTER_TIMEZONE	},	// 30
	{	5,	TYPE_RADIOBUTTON,	" +5.5",			NONE,					NONE,			 330,			CLUSTER_TIMEZONE	},	// 31
	{	5,	TYPE_RADIOBUTTON,	" +6",				NONE,					NONE,			 360,			CLUSTER_TIMEZONE	},	// 32
	{	5,	TYPE_RADIOBUTTON,	" +7",				NONE,					NONE,			 420,			CLUSTER_TIMEZONE	},	// 33
	{	5,	TYPE_RADIOBUTTON,	" +8",				NONE,					NONE,			 480,			CLUSTER_TIMEZONE	},	// 34
	{	5,	TYPE_RADIOBUTTON,	" +9",				NONE,					NONE,			 540,			CLUSTER_TIMEZONE	},	// 35
	{	5,	TYPE_RADIOBUTTON,	" +9.5",			NONE,					NONE,			 570,			CLUSTER_TIMEZONE	},	// 36
	{	5,	TYPE_RADIOBUTTON,	"+10",				NONE,					NONE,			 600,			CLUSTER_TIMEZONE	},	// 37
	{	5,	TYPE_RADIOBUTTON,	"+11",				NONE,					NONE,			 660,			CLUSTER_TIMEZONE	},	// 38
	{	5,	TYPE_RADIOBUTTON,	"+12",				NONE,					NONE,			 720,			CLUSTER_TIMEZONE	},	// 39
	{	6,	TYPE_ITEM,			"Exit",				IMAGE_ICON_EXIT,		CMD_EXIT,		NONE,			NONE				},	// 40
	{	6,	TYPE_CHECKBOX,		"Speedo (MPH)",		NONE,					NONE,			NONE,			GAUGE_SPEEDO_MPH	},	// 41
	{	6,	TYPE_CHECKBOX,		"Speedo (Km/h)",	NONE,					NONE,			NONE,			GAUGE_SPEEDO_KMPH	},	// 42
	{	6,	TYPE_CHECKBOX,		"Speedo (Knots)",	NONE,					NONE,			NONE,			GAUGE_SPEEDO_KNOTS	},	// 43
	{	6,	TYPE_CHECKBOX,		"Tacho",			NONE,					NONE,			NONE,			GAUGE_TACHO			},	// 44
	{	6,	TYPE_CHECKBOX,		"Clock",			NONE,					NONE,			NONE,			GAUGE_CLOCK			},	// 45
	{	6,	TYPE_CHECKBOX,		"Heading",			NONE,					NONE,			NONE,			GAUGE_HEADING		},	// 46
	{	6,	TYPE_CHECKBOX,		"GPS Home",			NONE,					NONE,			NONE,			GAUGE_GPS_HOME		},	// 47
	{	6,	TYPE_CHECKBOX,		"GPS Location",		NONE,					NONE,			NONE,			GAUGE_GPS_LOCATION	},	// 48
	{	7,	TYPE_ITEM,			"Exit",				IMAGE_ICON_EXIT,		CMD_EXIT,		NONE,			NONE				},	// 49
	{	7,	TYPE_RADIOBUTTON,	"Analog",			NONE,					NONE,			STYLE_ANALOG,	CLUSTER_STYLE		},	// 50
	{	7,	TYPE_RADIOBUTTON,	"Digital",			NONE,					NONE,			STYLE_DIGITAL,	CLUSTER_STYLE		},	// 51
	{	8,	TYPE_ITEM,			"Exit",				IMAGE_ICON_EXIT,		CMD_EXIT,		NONE,			NONE				},	// 52
	{	8,	TYPE_RADIOBUTTON,	"White",			NONE,					NONE,			0,				CLUSTER_COLOUR		},	// 53
	{	8,	TYPE_RADIOBUTTON,	"Red",				NONE,					NONE,			1,				CLUSTER_COLOUR		},	// 54
	{	8,	TYPE_RADIOBUTTON,	"Orange",			NONE,					NONE,			2,				CLUSTER_COLOUR		},	// 55
	{	8,	TYPE_RADIOBUTTON,	"Yellow",			NONE,					NONE,			3,				CLUSTER_COLOUR		},	// 56
	{	8,	TYPE_RADIOBUTTON,	"Green",			NONE,					NONE,			4,				CLUSTER_COLOUR		},	// 57
	{	8,	TYPE_RADIOBUTTON,	"Blue",				NONE,					NONE,			5,				CLUSTER_COLOUR		}	// 58
};


											
// +-------+
// | Fonts |
// +-------+

static const byte TEXT_ALIGN_LEFT	= 1;
static const byte TEXT_ALIGN_RIGHT	= 2;
static const byte TEXT_ALIGN_CENTER	= 3;

typedef struct FontType {
	char		firstChar;
	DWORD		address;		// Address of raw data on uSD.
	byte		characters;
	const byte*	characterWidths;// Array of widths
	byte		spaceWidth;		// px
	byte		letterSpacing;	// px
	byte		lineHeight;		// px
};


// Alternate Gothic
// ----------------

// XXXL (Digital Speedo)					     0   1   2   3   4   5   6   7   8   9
static const byte fontXXXLWidths[10]		= { 48, 48, 48, 48, 48, 48, 48, 48, 48, 48 };
static const FontType fontXXXL				= { '0', 0x00008000, 10, fontXXXLWidths, 48, 8, 104 };

// XXL (Digital RPM)						     0   1   2   3   4   5   6   7   8   9
static const byte fontXXLWidths[10]			= { 32, 32, 32, 32, 32, 32, 32, 32, 32, 32 };
static const FontType fontXXL				= { '0', 0x00020800, 10, fontXXLWidths, 32, 5, 80 };

// XL (Digital Clock)						     0   1   2   3   4   5   6   7   8   9  :
static const byte fontXLWidths[11]			= { 29, 17, 28, 27, 31, 27, 29, 26, 28, 29, 9 };
static const FontType fontXL				= { '0', 0x0002D200, 11, fontXLWidths, 16, 3, 64 };

// L (Digital GPS Location)						 !   "   #   $   %   &  '   (   )   *   +  ,   -  .   /   0   1   2   3   4   5   6   7   8   9  :  ;   <   =   >   ?   @   A   B   C   D   E   F   G   H  I   J   K   L   M   N   O   P   Q   R   S   T   U   V   W   X   Y   Z   [   \   ]      _ 
static const byte fontLWidths[63]			= {  6, 15, 24, 15, 26, 21, 6, 10, 10, 18, 14, 6, 14, 6, 18, 18, 11, 18, 18, 18, 18, 18, 18, 18, 18, 6, 6, 32, 14, 32, 15, 32, 18, 18, 18, 18, 18, 18, 18, 18, 6, 11, 20, 15, 24, 18, 18, 18, 20, 18, 18, 18, 18, 19, 28, 19, 19, 17, 12, 18, 12, 15, 18 };
static const FontType fontL					= { '!', 0x00038400, 63, fontLWidths, 6, 2, 48 };

// M (Analog GPS Location)						 !   "   #   $   %   &  '  (  )   *  +  ,  -  .   /   0  1   2   3   4   5   6   7   8   9  :  ;   <  =   >   ?   @   A   B   C   D   E   F   G   H  I  J   K   L   M   N   O   P   Q   R   S   T   U   V   W   X   Y   Z  [   \  ]      _ 
static const byte fontMWidths[63]			= {  4, 12, 32, 10, 17, 14, 6, 7, 7, 12, 9, 4, 9, 4, 12, 12, 7, 12, 12, 12, 12, 12, 12, 12, 12, 4, 4, 21, 9, 21, 10, 24, 12, 12, 12, 12, 12, 12, 12, 12, 4, 7, 13, 10, 16, 12, 12, 12, 13, 12, 12, 12, 12, 13, 19, 13, 13, 11, 8, 12, 8, 12, 12 };
static const FontType fontM					= { '!', 0x00067A00, 63, fontMWidths, 4, 1, 32 };



// Eurostyle LT Bold
// -----------------

// S (Digital Labels + AM/PM and Analog Home)	 !  "  #   $   %   &  '  (  )  *  +  ,  -  .   /   0  1  2  3   4  5  6  7  8  9  :  ;  <  =  >  ?   @   A   B   C   D  E   F  G   H  I  J   K  L   M   N   O   P   Q   R   S  T   U   V   W   X   Y   Z  [  \  ]    _ 
static const byte fontSWidths[63]			= {  2, 5, 9, 11, 14, 11, 2, 4, 4, 6, 8, 2, 4, 2, 11, 11, 7, 9, 9, 10, 9, 9, 9, 9, 9, 2, 2, 8, 8, 8, 8, 13, 11, 10, 10, 11, 8, 8, 10, 10, 2, 8, 10, 8, 14, 11, 10, 10, 11, 10, 10, 9, 10, 11, 16, 11, 11, 10, 4, 6, 4, 5, 8 };
static const FontType fontS_White			= { '!', 0x00087400, 63, fontSWidths, 4, 1, 16 };
static const FontType fontS_Red				= { '!', 0x0008F400, 63, fontSWidths, 4, 1, 16 };
static const FontType fontS_Orange			= { '!', 0x00097400, 63, fontSWidths, 4, 1, 16 };
static const FontType fontS_Yellow			= { '!', 0x0009F400, 63, fontSWidths, 4, 1, 16 };
static const FontType fontS_Green			= { '!', 0x000A7400, 63, fontSWidths, 4, 1, 16 };
static const FontType fontS_Blue			= { '!', 0x000AF400, 63, fontSWidths, 4, 1, 16 };


// Font Colours
typedef struct ColourType {
	const char	*name;
	word		rgb565;
	FontType	font;
};
static const ColourType colours[6] = {
	{ "White",	0xffff, fontS_White		},
	{ "Red",	0xe904, fontS_Red		},
	{ "Orange",	0xfb20, fontS_Orange	},
	{ "Yellow",	0xff00, fontS_Yellow	},
	{ "Green",	0x07e0, fontS_Green		},
	{ "Blue",	0x033f, fontS_Blue		}
};






// Gauge Class
// --------------------------------------------------------------------------------
class Gauge {
	
public:		
	// +-------------------+
	// | Class Constructor |
	// +-------------------+
	Gauge(Stream * virtualPort, int intBTN, void * parentCluster);

	// +-------------------+
	// | Public Attributes |
	// +-------------------+
	Goldelox_Serial_4DLib	Display;

	int						screen;
	boolean					screens[SCREENS];
	byte					mode;				// Screen or Menu mode?
	
	// Custom variable-width font (each character is a frame of an .AVI file on the uOLED-160-G2's SD card).
	FontType				font;
	byte					cursorX;
	byte					cursorY;


	// +------------------+
	// | Public Functions |
	// +------------------+
	void	begin			(void);
	void	loop			(void);
	void	refresh			(void);
	void	select			(void);
	void	ok				(void);

	// Font routines - I'll move these to Private after testing :)
	int		fontWidth		(char *strText);
	void	fontWrite		(char ch);
	void	fontPrint		(char *strText, byte x, byte y, byte align, boolean clearLeft, boolean clearRight);


private:

	// +--------------------+
	// | Private Attributes |
	// +--------------------+
	void							*cluster;
	
	// Button
	byte							buttonPin;
	byte							buttonWas;
	byte							buttonIs;
	unsigned long					buttonMillis;
	boolean							buttonHeld;		// Has the button been held?

	// Menu Attributes
	MenuItemType					items[MAX_ITEMS_PER_MENU];
	byte							itemCount;
	byte							parent;
	byte							offset;
	byte							cursor;

	// Misc attributes
	short							minLastDrawn;	// We check what the last minute was displayed in the loop. if it's different, we refresh the clock.
	boolean							screenInitialised;



	// +-------------------+
	// | Private Functions |
	// +-------------------+
	void	setGPSHome		(void);
	void	about			(void);

	void	setScreen		(byte bytScreen);
	void	drawScreen		(byte bytScreen);

	void	drawSpeedo		(char *strLabel, DWORD dwrdBackground, byte bytSpeed, byte bytSpeedPerFrame, byte bytLastFrame);
	void	drawTacho		(void);
	void	drawClock		(void);
	void	drawHeading		(void);
	void	drawGPSHome		(void);
	void	drawGPSLocation	(void);
	void	drawSatellite	(char *strLabel);

	void	setMenu			(int intParent);
	void	exit			(void);
	void	drawMenu		(void);
	void	drawItem		(MenuItemType objItem, byte y, boolean bHighlight);

	word	getSetting		(MenuItemType objItem);
	void	setSetting		(MenuItemType objItem, word value);
};



// GaugeCluster Class
// --------------------------------------------------------------------------------



class GaugeCluster {
	
public:
	// +-------------------+
	// | Class Constructor |
	// +-------------------+
	GaugeCluster();


	// +-------------------+
	// | Public Attributes |
	// +-------------------+
	Gauge			*gauges;
	int				gaugeCount;

	// Set these attributes in your sketch's loop.
	boolean			fix;
	word			heading;			// degrees
	byte			knots;
	word			rpm;
	float			latitude;			// ddmm.mmmm
	char			latitudeNS;			// 'N' or 'S'
	float			longitude;			// ddmm.mmmm
	char			longitudeEW;		// 'E' or 'W'
	float			latitudeHome;		// 2756.0833
	char			latitudeHomeNS;		// 'S'
	float			longitudeHome;		// 15325.7000
	char			longitudeHomeEW;	// 'E'
	float			altitude;
    int				hour;
    int				minute;
    int				seconds;
    int				year;
    int				month;
    int				day;

	// These attributes show the state of the cluster as it was on the last pass of the loop.
	// This allows the screens to only be refreshed when they need to be, rather than on every pass.
	boolean			fixWas;
	word			headingWas;
	byte			knotsWere;
	int				rpmWas;
	float			latitudeWas;
	char			latitudeNSWas;
	float			longitudeWas;
	char			longitudeEWWas;
	float			latitudeHomeWas;
	char			latitudeHomeNSWas;
	float			longitudeHomeWas;
	char			longitudeHomeEWWas;
	float			altitudeWas;
    int				hourWas;
    int				minuteWas;
    int				secondsWere;
    int				yearWas;
    int				monthWas;
    int				dayWas;

	boolean			speedoNeedsFix;
	boolean			headingNeedsFix;
	boolean			clockNeedsFix;

	// Satellite Animation
	byte			frame;			// For the satellite animation.
	byte			frameWas;
	long			frameTimer;		// Timer used to wait between frames.

	// Gauge Cluster Editable Settings
	word			style;
	byte			colour;
	word			timezone;


	// +------------------+
	// | Public Functions |
	// +------------------+
	void			begin				();
	void			loop				();
	void			addGauge			(Stream * virtualPort, int intBTN);
	void			refreshAllGauges	();

	void			loadSettings		();
	void			saveSettings		();

	// Fonts
	


private:

	// +--------------------+
	// | Private Attributes |
	// +--------------------+

	// +-------------------+
	// | Private Functions |
	// +-------------------+
};



// Misc Functions
byte	daysInMonth			(byte bytMonth, int intYear);
char	*ftoa				(char *a, double f, int precision);
double	decimalDegrees		(double dblDDDMMpointMMMM, char NSEW);
double	DegreesToRadians	(double dblDegrees);
double	RadiansToDegrees	(double dblRadians);


#endif