/*
use an NPN transistor to power the uOLED-160-G2:
Collector = Arduino 5V
Base      = Arduino IO pin (4)
Emitter   = uOLED-160-G2 5V
*/

#define PIN_DISPLAY_5V        4
#define PIN_BUTTON           12
#define PIN_RPM              A0
#define STARTUP_DELAY         3 // Seconds

#define GPS_MPH_PER_KNOT      0.51444444
#define GPS_KMPH_PER_KNOT     1.852

#define UPDATE_DELAY         100 // = 1000 / FPS


// Gauge Cluster
#include <avr/eeprom.h>        // For loading / Saving the settings
#include <SoftwareSerial.h>
#include <Goldelox_Serial_4DLib.h>
#include <GaugeCluster.h>

// GPS
#include <Adafruit_GPS.h>   // For interpreting the GPS data stream


// Specifically for the Arduino Mega 2560 (or 1280 on the original Arduino Mega)
// A majority of the pins are NOT PCINTs, SO BE WARNED (i.e. you cannot use them as receive pins)
// Only pins available for RECEIVE (TRANSMIT can be on any pin):
// (I've deliberately left out pin mapping to the Hardware USARTs - seems senseless to me)
// Pins: 10, 11, 12, 13,  50, 51, 52, 53,  62, 63, 64, 65, 66, 67, 68, 69
// So pins 2 & 3 won't work (for receive)
SoftwareSerial DisplaySerial(11, 10);
GaugeCluster Cluster;


// GPS parser
Adafruit_GPS GPS(&Serial1);



void setup() {
      
    Serial.begin(9600); // Serial Monitor (for debugging)
    
    setupDisplay();
    
    setupGPS();
    
    pinMode(PIN_RPM, INPUT); // For now, we'll fake RPM by reading from a potentiometer on an analog input.
}


void loop() {
  
    long lngMillis;
    
    char c = GPS.read();
    if (GPS.newNMEAreceived()) {
        if (GPS.parse(GPS.lastNMEA())) {
            
            Cluster.fix = GPS.fix;
            if (GPS.fix) {
                // Speed/Direction
                Cluster.knots       = GPS.speed;
            	Cluster.heading     = GPS.angle;
            
                // Position
            	Cluster.latitude    = GPS.latitude;  // ddmm.mmmm         2756.1354
                Cluster.latitudeNS  = GPS.lat;       // 'N' or 'S'              'S'
            	Cluster.longitude   = GPS.longitude; // ddddmm.mmmm      15325.2901
                Cluster.longitudeEW = GPS.lon;       // 'E' or 'W'              'E'
                
                // Time
                Cluster.hour        = GPS.hour;
                Cluster.minute      = GPS.minute;
                Cluster.seconds     = GPS.seconds;
                Cluster.year        = GPS.year;   // yyyy
                Cluster.month       = GPS.month;  // 1..12
                Cluster.day         = GPS.day;    // 1..31
            }
        }
    }
        
    // Analog Input (for now).
    Cluster.rpm      = map(analogRead(A0), 0, 1023, 0, 160) * 50;
    //Cluster.heading  = map(analogRead(A0), 0, 1023, 0, 359);
    
    // Only update the displays if we've waited long enough.
    if (millis() >= lngMillis) {
        lngMillis = millis() + UPDATE_DELAY;
        Cluster.loop();
    } 
}



void Callback(int ErrCode, unsigned char ErrByte) {

    const char *Error4DText[] = { "OK\0", "Timeout\0", "NAK\0", "Length\0", "Invalid\0" };
    
    Serial.print(F("Serial 4D Library reports error "));
    Serial.print(Error4DText[ErrCode]);
    
    if (ErrCode == Err4D_NAK) {
        Serial.print(F(" returned data = "));
        Serial.println(ErrByte);
    } else Serial.println(F(""));
}



void setupDisplay() {
    
    DisplaySerial.begin(9600); // 4D Systems UOLED-160-G2 Display

    pinMode(PIN_DISPLAY_5V, OUTPUT);
    digitalWrite(PIN_DISPLAY_5V, HIGH);
    
    // The uOLED-160-G2 needs about 3 seconds to initialise before you can send commands.
    Serial.print("Waiting for uOLED");
    for (int i = 0; i < STARTUP_DELAY; i++) {
        Serial.print(".");
        delay(1000);
    }
    Serial.println("OK.");    
        
    // Add the gauges to the Cluster object.
    Cluster.addGauge(&DisplaySerial, PIN_BUTTON);            
    
    // Display Error Handling
    Cluster.gauges[0].Display.TimeLimit4D = 2000;      // 2 second timeout on all commands
    Cluster.gauges[0].Display.Callback4D  = Callback ; // Without this, you'll get "Internal Function Missing" errors on the display.
        
    // Send the command to set the baud rate at the display end.
    // SoftwareSerial can only go as fast as 38400.
    // If we use a hardware serial port we can go as fast as 115200.
    DisplaySerial.print((char)(F_setbaudWait >> 8));
    DisplaySerial.print((char)(F_setbaudWait));
    DisplaySerial.print((char)(BAUD_38400 >> 8));
    DisplaySerial.print((char)(BAUD_38400));

    // Change the baud rate at the Arduino end
    DisplaySerial.flush();
    DisplaySerial.end();
    delay(50);
    DisplaySerial.begin(38400);
    
    // Get the ACK
    Cluster.gauges[0].Display.GetAck();
    
    // Go!!
    Cluster.begin();
}



void setupGPS() {
  
    Serial1.begin(9600); // Adafruit Ultimate GPS Shield (with jumper-wire from shield pin 8 to MEGA pin 19)

    // Set up the GPS
    GPS.sendCommand(PMTK_SET_NMEA_OUTPUT_RMCGGA); // Recommended Minimum and fix data
    GPS.sendCommand(PMTK_SET_NMEA_UPDATE_5HZ);    // 1Hz, 5Hz, or 10Hz.
    GPS.sendCommand(PGCMD_ANTENNA);               // Receive updates about antenna
}
